#include "algebra.h"

using namespace cv;

namespace ltflo {


using std::vector;

/* determinant of square matrix 3x3 */
double det3(const double * A) {
	return A[0]*A[4]*A[8] + A[1]*A[5]*A[6] + A[3]*A[7]*A[2] - A[2]*A[4]*A[6] - A[7]*A[5]*A[0] - A[3]*A[1]*A[8];
}

/* In-place inversion of 3x3 real square matrix. Returns 0 on success, 1 on bad conditioning, 2 when singular; */
char inv3(double *a) {
	double ai[3*3]; /*inverted*/
	double det = det3(a); /*a[0]*a[4]*a[8]-a[0]*a[5]*a[7]-a[1]*a[3]*a[8]+a[1]*a[5]*a[6]+a[2]*a[3]*a[7]-a[2]*a[4]*a[6];*/
	double EPS = 1e-16;
	
	if (det == 0) {
		return 2;
	}
	
	ai[0] = (a[4]*a[8]-a[5]*a[7]) / det;
	ai[1] = (a[2]*a[7]-a[1]*a[8]) / det;
	ai[2] = (a[1]*a[5]-a[2]*a[4]) / det;
	ai[3] = (a[5]*a[6]-a[3]*a[8]) / det;
	ai[4] = (a[0]*a[8]-a[2]*a[6]) / det;
	ai[5] = (a[2]*a[3]-a[0]*a[5]) / det;
	ai[6] = (a[3]*a[7]-a[4]*a[6]) / det;
	ai[7] = (a[1]*a[6]-a[0]*a[7]) / det;
	ai[8] = (a[0]*a[4]-a[1]*a[3]) / det;
	
	memcpy(a, ai, 3*3*sizeof(double));
	
	if (det < EPS) {
		return 1;
	} else {
		return 0;
	}
}

/* rotation from Affine transformation (also Similarity and others) */
double a2rot(const double *A) {
	return asin(A[1] / sqrt(A[0] * A[4] - A[1] * A[3]));
}
double a2rot(const cv::Mat& A) {
	return asin(A.at<double>(1,0) / sqrt(A.at<double>(0,0) * A.at<double>(1,1) - A.at<double>(1,0) * A.at<double>(0,1)));
}

/*get scale change of any transf. matrix up to affine*/
double getScaleChange(double * A) {
	return sqrt(A[0]*A[4] - A[1]*A[3]);
}
double getScaleChange(const Mat& A) {
	return sqrt(A.at<double>(0,0)*A.at<double>(1,1) - A.at<double>(1,0)*A.at<double>(0,1));
}

/*get rotation of any transf. matrix up to affine*/
double getRotation(double * A) {
	return asin(A[2] / getScaleChange(A)); /*because A[2,1] = S.sin(r) for similarity and similar for others*/
}

/*random uniform double from interval (0..1]*/
double drand() {
	return (rand()+1.0)/(RAND_MAX+1.0);
}

/*random double with normal distribution with mu = 0 and sigma = 1*/
double randNorm() {
	return sqrt(-2*log(drand())) * cos(2*M_PI*drand());
}

/*Cross-product of lines or points from u*/
void crossU(double * p, const double * u, unsigned i1, unsigned i2, unsigned iu, unsigned u_len) {
	unsigned b1 = i1 * u_len + iu * 3, b2 = i2 * u_len + iu * 3; /*bases of vectors*/
	p[0] = u[b1 + 1] * u[b2 + 2] - u[b1 + 2] * u[b2 + 1];
	p[1] = u[b1 + 2] * u[b2 + 0] - u[b1 + 0] * u[b2 + 2];
	p[2] = u[b1 + 0] * u[b2 + 1] - u[b1 + 1] * u[b2 + 0];
}

/* Cross-product of two general 3-vectors */
/* X = AxB */
void cross(double *X, const double *A, const double *B) {
	X[0] = A[1]*B[2] - A[2]*B[1];
	X[1] = A[2]*B[0] - A[0]*B[2];
	X[2] = A[0]*B[1] - A[1]*B[0];
}


/*Least Squares solution to A.x=b by pseudoinverse, A is matrix MxN stored columnwise*/
int lsqPInv(double * x, double * A, double * b, unsigned M, unsigned N) {
	/*pseudoinverse: x = (A^T A)^-1 A^T b minimises ||Ax - b||*/
	double * Ant; /*not transposed, A taken as already transposed!*/
	double * AtA, * Apinv;
	int res;
	
	/*CCMath uses row-wise storage!!!*/
	Ant = (double *) malloc(M * N * sizeof(double));
	mattr(Ant, A, N, M);
	
	AtA = (double *) malloc(N * N * sizeof(double));
	rmmult(AtA, A, Ant, N, M, N);
	res = minv(AtA, N); /*AtA now inverted*/
	
	Apinv = Ant; /*now used as pseudoinverse of A*/
	rmmult(Apinv, AtA, A, N, N, M);
	
	/*result, multiply right side by pseudoinverse*/
	rmmult(x, Apinv, b, N, M, 1);
	
	/*CCMath uses row-wise storage!!!*/
	
	free(AtA);
	free(Ant);
	return res;
}

/*iterative non-linear minimisation of geometric error of inliing correspondences*/
/*void l2aNL(double * A, const double * u, unsigned I, const unsigned char * inliers, const double * win, unsigned len, unsigned u_len) {
	double v[4];
	int iter;
	double * l1, * l2, *d, * k1, * k2;
	unsigned i;
	arr6 arr;
	
	k1 = l1 = (double *) malloc(3 * I * sizeof(double));
	k2 = l2 = (double *) malloc(3 * I * sizeof(double));
	d = (double *) malloc(I * sizeof(double));
	
	for (i = 0; i < len; ++i) {
		if (*inliers++) {
			memcpy(l1, u, 3 * sizeof(double));
			memcpy(l2, u+3, 3 * sizeof(double));
			l1 += 3;
			l2 += 3;
		}
		u += u_len;
	}
	arr.l1 = k1;
	arr.l2 = k2;
	arr.d = d;
	arr.win = win;
	arr.size = I;
	
	v[0] = A[0] * 100;
	v[1] = A[1] * 100;
	v[2] = A[6];
	v[3] = A[7];
	
	iter = dlevmar_dif(&geomError, v, NULL, 4, 1, 1000, NULL, NULL, NULL, NULL, (void *) &arr);
	
	A[0] = A[4] = v[0] / 100;
	A[1] = v[1] / 100;
	A[2] = A[5] = 0;
	A[3] = -v[1] / 100;
	A[6] = v[2];
	A[7] = v[3];
	A[8] = 1;
	
	free(k1);
	free(k2);
	free(d);
}*/

///*iterative non-linear minimisation of geometric error of inliing correspondences using the Nelder-Mead algorithm*/
//void l2aNLnm(double * A, const double * u, unsigned I, const bool * inliers, const double * win, unsigned len, unsigned u_len, double *buffer) {
//	double vs[4], v[4];
//	double * l1, * l2, *d;
////	double * u1, * u2;
//	unsigned i;
//	arr6 arr;
//	double minimum;
//	double fstep[] = {10, 10, 10, 10};
//	int nEv, nRes, fIn;
//	double scaling = 100;
//	
//	l1 = (double *) malloc(3 * I * sizeof(double));
//	l2 = (double *) malloc(3 * I * sizeof(double));
//	d = (double *) malloc(I * sizeof(double));
////	u1 = (double *) malloc(3 * I * sizeof(double));
////	u2 = (double *) malloc(3 * I * sizeof(double));
//	
//	for (i = 0; i < len; ++i) {
//		if (*inliers++) {
//			memcpy(l1, u, 3 * sizeof(double));
//			memcpy(l2, u+3, 3 * sizeof(double));
//			l1 += 3;
//			l2 += 3;
////			memcpy(u1, u+6, 3 * sizeof(double));
////			memcpy(u2, u+9, 3 * sizeof(double));
////			u1 += 3;
////			u2 += 3;
//		}
//		u += u_len;
//	}
//	l1 -= 3 * I; /* back to the beginning */
//	l2 -= 3 * I;
////	u1 -= 3 * I;
////	u2 -= 3 * I;
//	arr.l1 = l1;
//	arr.l2 = l2;
////	arr.u1 = u1;
////	arr.u2 = u2;
//	arr.d = d;
//	arr.win = win;
//	arr.size = I;
//	arr.buffer = buffer;
//	
//	vs[0] = A[0] * scaling;
//	vs[1] = A[1] * scaling;
//	vs[2] = A[6];
//	vs[3] = A[7];
//	
////	printf("Geom. error before optimisation: %10.1f\n", geomErrorSimple4D(vs, &arr));
//	nelmin ( geomErrorSimple, 4, vs, v, &minimum, .1, fstep, 10, 500, &nEv, &nRes, &fIn, &arr);
////	nelmin ( geomErrorSimple4D, 4, vs, v, &minimum, .1, fstep, 10, 500, &nEv, &nRes, &fIn, &arr);
////	printf("Geom. error after optimisation:  %10.1f, %d evals used, %dx restarted\n", geomErrorSimple4D(v, &arr), nEv, nRes);
//	
//	A[0] = A[4] = v[0] / scaling;
//	A[1] = v[1] / scaling;
//	A[2] = A[5] = 0;
//	A[3] = -v[1] / scaling;
//	A[6] = v[2];
//	A[7] = v[3];
//	A[8] = 1;
//	
//	free(l1);
//	free(l2);
////	free(u1);
////	free(u2);
//	free(d);
//	
//}

///* linearised step minimising orig. geometric error of inliing correspondences using the Simon's magic
//	here the buffer must be at least 12*size */
//void l2aSimonOrig(double * A, const double * u, unsigned I, const unsigned * inliers, const double * win, unsigned u_len, double *buffer) {
//	std::vector<double> vs(4), v;
//	
//	/* scale*cos(a), scale*sin(a), dx, dy */
//	vs[0] = A[0];
//	vs[1] = A[1];
//	vs[2] = A[6];
//	vs[3] = A[7];
//	
//	fw::Linear_taylor_solver solver;
////	vector<fw::Opt_func *> funcs(1);
////	funcs[0] = (fw::Opt_func *) new LineErrorOrigF(u, win, I, inliers, u_len, buffer);
//	fw::Opt_func * opt_fun = (fw::Opt_func *) new LineErrorOrigF(u, win, I, inliers, u_len, buffer);
//	vector<fw::Grad_func *> grads(1);
//	grads[0] = (fw::Grad_func *) new LineErrorOrigG(((LineErrorOrigF *)opt_fun)->getL1(), ((LineErrorOrigF *)opt_fun)->getL2(), win, I, opt_fun);
//	solver.set_grad_funcs(grads);
//	vector<vector<double> > tmp(1, vector<double>(4));
////	tmp[0][0] = tmp[0][1] = tmp[0][2] = tmp[0][3] = 1e-8;
////	solver.set_finite_difference_steps(tmp);
//	solver.set_max_iters(2);
//	solver.set_num_dimensions(4);
//	solver.set_num_equations(4*I);
//	
//	v = solver.optimise(vs);
////	vector<double> res0 = (*funcs[0])(vs);
////	vector<double> res1 = (*funcs[0])(v);
////	printf("Error before opt.: %.2f, error after opt.: %.2f\n", fw::sum_squares(res0), fw::sum_squares(res1));
////	using namespace fw;
////	cout << "before: " << vs << ", after: " << v << '\n';
////	cout << "first few BEFORE: " << res0[0] << ", " << res0[1] << ", " << res0[2] << ", " << res0[3] << '\n';
////	cout << "first few AFTER: " << res1[0] << ", " << res1[1] << ", " << res1[2] << ", " << res1[3] << '\n';
//	
//	A[0] = A[4] = v[0];
//	A[1] = v[1];
//	A[2] = A[5] = 0;
//	A[3] = -v[1];
//	A[6] = v[2];
//	A[7] = v[3];
//	A[8] = 1;
//}

///* linearised step minimising geometric error of inliing correspondences using the Simon's magic
//	here the buffer must be at least 24*size */
//void l2aSimon(double * A, const double * u, unsigned I, const unsigned * inliers, const double * win, unsigned u_len, double *buffer) {
//	std::vector<double> vs(4), v;
//	
//	/* scale*cos(a), scale*sin(a), dx, dy */
//	vs[0] = A[0];
//	vs[1] = A[1];
//	vs[2] = A[6];
//	vs[3] = A[7];
//	
//	fw::Linear_taylor_solver solver;
//	vector<fw::Opt_func *> funcs(1);
//	funcs[0] = (fw::Opt_func *) new LineErrorF(u, I, inliers, u_len, buffer);
//	solver.set_funcs(funcs);
//	vector<vector<double> > tmp(1, vector<double>(4));
//	tmp[0][0] = tmp[0][1] = tmp[0][2] = tmp[0][3] = 1e-5;
//	solver.set_finite_difference_steps(tmp);
//	solver.set_max_iters(2);
//	solver.set_num_dimensions(4);
//	solver.set_num_equations(4*I);
//	
//	v = solver.optimise(vs);
////	vector<double> res0 = (*funcs[0])(vs);
////	vector<double> res1 = (*funcs[0])(v);
////	printf("Error before opt.: %.2f, error after opt.: %.2f\n", fw::sum_squares(res0), fw::sum_squares(res1));
////	using namespace fw;
////	cout << "before: " << vs << ", after: " << v << '\n';
////	cout << "first few BEFORE: " << res0[0] << ", " << res0[1] << ", " << res0[2] << ", " << res0[3] << '\n';
////	cout << "first few AFTER: " << res1[0] << ", " << res1[1] << ", " << res1[2] << ", " << res1[3] << '\n';
//	
//	A[0] = A[4] = v[0];
//	A[1] = v[1];
//	A[2] = A[5] = 0;
//	A[3] = -v[1];
//	A[6] = v[2];
//	A[7] = v[3];
//	A[8] = 1;
//}

///* linearised step maximising image evidence directly using the Simon's magic */
//void l2aEvid(double * A, const double * u, unsigned I, const unsigned * inliers, const double * evidDist, const double * evidAng, unsigned M, unsigned N, unsigned u_len, bool repair) {
//	std::vector<double> vs(4), v;
//	
//	/* scale*cos(a), scale*sin(a), dx, dy */
//	vs[0] = A[0];
//	vs[1] = A[1];
//	vs[2] = A[6];
//	vs[3] = A[7];
//	
//	fw::Linear_taylor_solver solver;
//	vector<fw::Opt_func *> funcs(1);
//	funcs[0] = (fw::Opt_func *) new EvidF(u, I, inliers, u_len, evidDist, evidAng, M, N, repair);
//	solver.set_funcs(funcs);
//	vector<vector<double> > tmp(1, vector<double>(4));
//	tmp[0][0] = tmp[0][1] = tmp[0][2] = tmp[0][3] = 1e-8;
//	solver.set_finite_difference_steps(tmp);
//	solver.set_max_iters(2);
//	solver.set_num_dimensions(4);
//	solver.set_num_equations(2*I);
//	
//	v = solver.optimise(vs);
//	
//	
//	A[0] = A[4] = v[0];
//	A[1] = v[1];
//	A[2] = A[5] = 0;
//	A[3] = -v[1];
//	A[6] = v[2];
//	A[7] = v[3];
//	A[8] = 1;
//}

/* function [A, score] = refineA(A, u2, weightsField) */
/* refines A to maximize fit of points in u2 backprojected by A^-1 to weightsField */
/*
Returns:
final score
Parameters:
A            I/O   3x3    similarity matrix to refine (from f_t-1 to f_t)
u2            I   2xLEN   points in the following frame (f_t)
weightsField  I    MxN    contour estimate (smooth) in the previous frame (f_t-1)
M, N, LEN     I   scalar  image dimensions, number of points
*/
void refineA(double *A, double * score, const double *u2, const double *weightsField, unsigned M, unsigned N, unsigned LEN) {
	double v[4], vs[4];
	arr5 arr;
	double minimum;
	double fstep[] = {.1, .1, 1, 1};
	int nEv, nRes, fIn;
	
	arr.wf = weightsField;
	arr.u2 = u2;
	arr.M = M;
	arr.N = N;
	arr.LEN = LEN;
	
	vs[0] = A[0];
	vs[1] = A[1];
	vs[2] = A[6];
	vs[3] = A[7];
	
/*	printf("Score before optimisation: %10.4f\n", -refineFitError(vs, &arr));*/
	throw "nelmin ( refineFitError, 4, vs, v, &minimum, .00001, fstep, 10, 1000, &nEv, &nRes, &fIn, &arr)";
/*	printf("Score after  optimisation: %10.4f (%.4f), %d evals used, %dx restarted\n", -refineFitError(v, &arr), -minimum, nEv, nRes);*/
	
	A[0] = A[4] = v[0];
	A[1] = v[1];
	A[2] = A[5] = 0;
	A[3] = -v[1];
	A[6] = v[2];
	A[7] = v[3];
	A[8] = 1;
	
	*score = -minimum;
	
}
// the same thing using OpenCV
void refineA(cv::Mat& A, double* score, const cv::Mat& u2, const cv::Mat& weightsField) {
	cv::Ptr<cv::DownhillSolver> solver = cv::DownhillSolver::create();
	cv::Ptr<cv::MinProblemSolver::Function> objective = cv::makePtr<WeightsF>(weightsField, u2);
	
	Mat v(1, 4, CV_64F);
	v.at<double>(0,0) = A.at<double>(0,0);
	v.at<double>(0,1) = A.at<double>(1,0);
	v.at<double>(0,2) = A.at<double>(0,2);
	v.at<double>(0,3) = A.at<double>(1,2);
	Mat steps = (Mat_<double>(1, 4) << 0.1, 0.1, 10, 10);
	
	solver->setFunction(objective);
	solver->setInitStep(steps);
	
	try
	{
	//	std::cout << "Opt score before: " << -objective->calc((double*)v.ptr()) << "\n";
		*score = - solver->minimize(v);
	//	std::cout << "Opt score after : " << -objective->calc((double*)v.ptr()) << "\n";
	
		A = Mat::eye(3, 3, CV_64F);
		A.at<double>(0, 0) = A.at<double>(1, 1) = v.at<double>(0,0);
		A.at<double>(1, 0) = v.at<double>(0,1);
		A.at<double>(0, 1) = -v.at<double>(0,1);
		A.at<double>(0, 2) = v.at<double>(0,2);
		A.at<double>(1, 2) = v.at<double>(0,3);
	}
	catch(...)
	{
		std::cerr<<"opencv exception thrown during refinement. Leaving unrefined"<<std::endl;
	}
}



/* Determinant of a square positive-definite matrix D×D via eigenvalues */
double determinantSym(const double * A, const unsigned D) {
	double det = 1;
	double * eig = (double *) malloc (D * sizeof(double) ); // eigenvalues
	unsigned i;
	// the matrix is altered during the computation
	double * mat = (double *) malloc ( D*D * sizeof(double) );
	memcpy(mat, A, D*D * sizeof(double) );
	eigval(mat, eig, D);
	for ( i = 0; i < D; ++i ) {
		det *= eig[i];
//		printf("%5.1f", eig[i]);
	}
//	printf("\n\n");
	
	free(mat);
	free(eig);
	return det;
}

double normpdfN(const double *x, const double *mu, const double *sigma, const unsigned D) {
	double dist[D], sigmaId[D];
	double sigmaI[D*D];
	
	for (unsigned d = 0; d < D; ++d) {
		dist[d] = x[d] - mu[d];
	}
	
	double sDet = determinantSym(sigma, D);
	if (sDet < 0) {
		std::cerr << "Negative-determinant covariance matrix encoutered!\n";
	}
	
	memcpy(sigmaI, sigma, D*D*sizeof(double));
	minv(sigmaI, D); // CCMath takes the input in row-major order, but it's symmetric, so it doesn't matter! (the same below)
	
	vmul(sigmaId, sigmaI, dist, D); // sigmaI * d
	double mahalanobis = vnrm(sigmaId, dist, D);
	
	return exp(- mahalanobis / 2) / sqrt(pow(2 * M_PI, D) * sDet);
}

// update multivariate normal distribution online, based on the results so far an the N-th observation according to
//   J. Bennett, R. Grout, P. Pebay, D. Roe, and D. Thompson. Numerically stable, single-pass, parallel statistics algorithms. In Cluster, 2009.
void updateMVNPD(double* mu, double* sigma, const double* x, unsigned N, unsigned D) {
	for (unsigned i = 0; i < D; ++i) {
		for (unsigned j = i; j < D; ++j) {
			sigma[i + j*D] = (sigma[i + j*D] * (N-1.0) + (x[i] - mu[i]) * (x[j] - mu[j]) * (N-1.0)/N) / N;
			// to be completely correct, here should be n-1 and on the first line n-2! This gives the initialisation/first observation larger weight, but avoids some init. problems...
			sigma[j + i*D] = sigma[i + j*D]; // symmetric
		}
		mu[i] += (x[i] - mu[i]) / N;
	}
}



} // namespace ltflo


