#ifndef __ALGEBRA_H__
#define __ALGEBRA_H__

#include <cstring>
/*#include <mex.h>*/
#include <cstdio>

#include "opencv2/opencv.hpp"


#include "ccmath.h"
//#include "levmar.h"
//#include "asa047.h"
//#include "solver.h"
//#include "linear_taylor_solver.h"

#include "cvutils.h"
#include "geometry.h"

namespace ltflo {

/*determinant of square matrix 3*3*/
double det3(const double * A);

/* In-place inversion of 3x3 real square matrix. Returns 0 on success, 1 on bad conditioning, 2 when singular; */
char inv3(double *A);

/* rotation from Affine transformation (also Similarity and others) */
double a2rot(const double *A);
double a2rot(const cv::Mat& A);

/*get scale change of any transf. matrix up to affine*/
double getScaleChange(double * A);
double getScaleChange(const cv::Mat& A);

/*get rotation of any transf. matrix up to affine*/
double getRotation(double * A);

/*random uniform double from interval (0..1]*/
double drand();

/*random normal double with mu = 0 and sigma = 1*/
double randNorm();

/*Cross-product of lines or points from u*/
void crossU(double * p, const double * u, unsigned i1, unsigned i2, unsigned iu, unsigned u_len);

/* Cross-product of two general 3-vectors */
/* X = AxB */
void cross(double *X, const double *A, const double *B);

/*Least Squares solution to A.x=b by pseudoinverse, A is matrix MxN stored columnwise*/
int lsqPInv(double * x, double * A, double * b, unsigned M, unsigned N);

/*iterative non-linear minimisation of geometric error of inliing correspondences*/
void l2aNL(double * A, const double * u, unsigned I, const unsigned char * inliers, const double * win, unsigned len, unsigned u_len);
/*iterative non-linear minimisation of geometric error of inliing correspondences using the Nelder-Mead algorithm*/
void l2aNLnm(double * A, const double * u, unsigned I, const bool * inliers, const double * win, unsigned len, unsigned u_len, double *buffer);
/* linearised step minimising geometric error of inliing correspondences using the Simon's magic
	here the buffer must be at least 24*size */
void l2aSimon(double * A, const double * u, unsigned I, const unsigned * inliers, const double * win, unsigned u_len, double *buffer);
/* linearised step minimising orig. geometric error of inliing correspondences using the Simon's magic
	here the buffer must be at least 12*size */
void l2aSimonOrig(double * A, const double * u, unsigned I, const unsigned * inliers, const double * win, unsigned u_len, double *buffer);
/* linearised step maximising image evidence directly using the Simon's magic */
void l2aEvid(double * A, const double * u, unsigned I, const unsigned * inliers, const double * evidDist, const double * evidAng, unsigned M, unsigned N, unsigned u_len, bool repair);
// estimates an affine transformation from a bbox correspondence or similar container of points
template <typename VecOfPts>
cv::Mat u2a(const VecOfPts& b1, const VecOfPts& b2, char type = 'a');
// estimates a similarity transformation from a bbox correspondence or similar container of points
template <typename VecOfPts>
cv::Mat u2s(const VecOfPts& b1, const VecOfPts& b2);


/* function [A, score] = refineA(A, u2, weightsField) */
/* refines A to maximize fit of points in u2 backprojected by A^-1 to weightsField */
/*
Returns:
final score
Parameters:
A            I/O   3x3    similarity matrix to refine (from f_t-1 to f_t)
u2            I   2xLEN   points in the following frame (f_t)
weightsField  I    MxN    contour estimate (smooth) in the previous frame (f_t-1)
M, N, LEN     I   scalar  image dimensions, number of points
*/
void refineA(double *A, double * score, const double *u2, const double *weightsField, unsigned M, unsigned N, unsigned LEN);
// the same thing using OpenCV
void refineA(cv::Mat& A, double* score, const cv::Mat& u2, const cv::Mat& weightsField);

/* Determinant of a symmetric square matrix D×D via eigen decomposition */
double determinantSym(const double * A, const unsigned D);

// multivariate normal distribution pdf
double normpdfN(const double *x, const double *mu, const double *sigma, const unsigned D);

// update multivariate normal distribution online, based on the results so far an the N-th observation
void updateMVNPD(double* mu, double* sigma, const double* x, unsigned N, unsigned D);














// estimates an affine transformation from a bbox correspondence
// the input is a vector of opencv points or something similar
template <typename VecOfPts>
cv::Mat u2a(const VecOfPts& b1, const VecOfPts& b2, char type) {
	if (type == 's') {
		return u2s(b1, b2);
	} else if (type != 'a') {
		throw "Error: Unknown transform type ('a' and 's' accepted)!";
	}
	
	cv::Mat eqL = cv::Mat::zeros(2*b1.size(), 6, CV_64F);
	cv::Mat eqR = cv::Mat::zeros(2*b1.size(), 1, CV_64F);
	for (unsigned p = 0; p < b1.size(); ++p) {
		// right-hand side
		eqR.at<double>(p*2+0, 0) = b2[p].x;
		eqR.at<double>(p*2+1, 0) = b2[p].y;
		// left-hand side
		eqL.at<double>(p*2+0, 0) = b1[p].x;
		eqL.at<double>(p*2+0, 1) = b1[p].y;
		eqL.at<double>(p*2+0, 2) = 1;
		eqL.at<double>(p*2+1, 3) = b1[p].x;
		eqL.at<double>(p*2+1, 4) = b1[p].y;
		eqL.at<double>(p*2+1, 5) = 1;
	}
	
	// now we want pseudoinverse: eqL\eqR
	cv::Mat solution = eqL.inv(cv::DECOMP_SVD) * eqR;
	// and finally fill the matrix
	cv::Mat A = cv::Mat::eye(3, 3, CV_64F);
	cv::Mat row1 = solution(cv::Rect(0,0,1,3)).t();
	cv::Mat row2 = solution(cv::Rect(0,3,1,3)).t();
	row1.copyTo(A.row(0));
	row2.copyTo(A.row(1));
	return A;
}
// estimates a similarity transformation from a bbox correspondence
// the input is a vector of opencv points or something similar
template <typename VecOfPts>
cv::Mat u2s(const VecOfPts& b1, const VecOfPts& b2) {
	throw "Error: Not implemented yet!";
}























} // namespace ltflo

#endif // __ALGEBRA_H__


